package java.util.concurrent.atomic;

import java.util.function.BinaryOperator;
import java.util.function.UnaryOperator;
import sun.misc.Unsafe;
import java.io.Serializable;

public class AtomicReference<V> implements Serializable
{
    private static final long serialVersionUID = -1848883965231344442L;
    private static final Unsafe unsafe;
    private static final long valueOffset;
    private volatile V value;
    
    public AtomicReference(V value) {
        this.value = value;
    }
    
    public AtomicReference() {
    }
    
    public final V get() {
        return this.value;
    }
    
    public final void set(V value) {
        this.value = value;
    }
    
    public final void lazySet(V x) {
        AtomicReference.unsafe.putOrderedObject(this, AtomicReference.valueOffset, x);
    }
    
    public final boolean compareAndSet(V expected, V x) {
        return AtomicReference.unsafe.compareAndSwapObject(this, AtomicReference.valueOffset, expected, x);
    }
    
    public final boolean weakCompareAndSet(V expected, V x) {
        return AtomicReference.unsafe.compareAndSwapObject(this, AtomicReference.valueOffset, expected, x);
    }
    
    public final V getAndSet(V newValue) {
        return AtomicReference.unsafe.getAndSetObject(this, AtomicReference.valueOffset, newValue);
    }
    
    public final V getAndUpdate(UnaryOperator<V> unaryOperator) {
        Object value;
        do {
            value = this.get();
        } while (!this.compareAndSet(value, unaryOperator.apply(value)));
        return value;
    }
    
    public final V updateAndGet(UnaryOperator<V> unaryOperator) {
        Object value;
        Object apply;
        do {
            value = this.get();
            apply = unaryOperator.apply(value);
        } while (!this.compareAndSet(value, apply));
        return apply;
    }
    
    public final V getAndAccumulate(V v, BinaryOperator<V> binaryOperator) {
        Object value;
        do {
            value = this.get();
        } while (!this.compareAndSet(value, binaryOperator.apply(value, v)));
        return value;
    }
    
    public final V accumulateAndGet(V v, BinaryOperator<V> binaryOperator) {
        Object value;
        Object apply;
        do {
            value = this.get();
            apply = binaryOperator.apply(value, v);
        } while (!this.compareAndSet(value, apply));
        return apply;
    }
    
    @Override
    public String toString() {
        return String.valueOf(this.get());
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        try {
            valueOffset = AtomicReference.unsafe.objectFieldOffset(AtomicReference.class.getDeclaredField("value"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
}
